/*
 * Add a property to an XMP file, or query the value of a property in
 * an XMP file.
 *
 * An XMP file is a file with RDF data in the RDF subset defined by
 * Adobe XMP.
 *
 * XMP is normally embedded in other files (such as JPEG, TIFF or PDF
 * files) to add metadata to them. This program does *not* handle
 * embedded XMP. Use a program such as rdjpgxmp or rdjpgapp to extract
 * the XMP from JPEG/JFIF and wrjpgapp to put it back in. Several
 * Adobe programs (such as Adobe Illustrator) can also extract XMP
 * from image formats supported by Adobe into a separate file and copy
 * XMP from an external file into an image file.
 *
 * Usage:
 * xmptool [-l language] property [file]
 * xmptool -w [-l language] property value [file]
 * xmptool -d [-l language] property [file]
 * xmptool -s [file]
 * xmptool -c [file]
 *
 * The default action is to look for the given property in the given
 * file (or stdin) and print its value. If the property has multiple
 * values, the default value is printed, or, if there is none, the
 * first value.
 *
 * The property must be a URI. If it is not, it is assumed to be a
 * Dublin Core property. (I.e., specifying "title" is equivalent to
 * specifying "http://purl.org/dc/elements/1.1/title".)
 *
 * -l language: print only the value that is in the given language
 *
 * -w (write) copies the given file (or stdin) to stdout with the
 *    given value added as a value for the given property. If -l is
 *    present, the value is marked as being in that language. Language
 *    must be an RFC xxxx value, e.g., "en" or "fr-ca". The file must
 *    be valid XMP or empty. Use /dev/null as file to create a new XMP
 *    file from scratch.
 *
 * -r (Only in combination with -w.) The value is to be interpreted as
 *    a URI (i.e., as an rdf:resource) instead of a string literal.
 *
 * -d (delete) deletes the given property. With -l, only deletes a
 *    value in that language and leaves values in other languages.
 *
 * -s Output the XMP file as a series of Bash shell arrays.
 *
 * -c Clean-up: remove values that are duplicates of default values
 *    (i.e., with language "x-default" or an empty language)
 *
 * Created: 7 August 2006
 * Author: Bert Bos <bert@w3.org>
 */

#include <config.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#include <ctype.h>
#include <fnmatch.h>
#include "heap.e"
#include "types.e"
#include "errexit.e"
#include "scan.e"
#include "html.e"

#define DC "http://purl.org/dc/elements/1.1/"

/* The symbol table is a chain of prefix/uri pairs. Every time an
 * element starts, the prefixes defined by it are added at the end. To
 * expand a prefix, the most recently added prefix/uri pair is used.
 * When en element ends, the chain is reduced to what it was when the
 * element started. The stack keeps track of where the chain ended at
 * the start of the element.
 *
 * ToDo: should we hash the prefixes? or is linear search good enough?
 **/
typedef struct _Symbol {
  string prefix;
  string uri;
  struct _Symbol *next;
} Symbol, *SymbolTable;

typedef struct _StackElt {
  Symbol *frame;
  struct _StackElt *next;
} *Stack;

typedef enum {SINGLE, ALT, BAG, SEQ} Multitype;

typedef struct _Property {
  string name;
  string val;
  string lang;
  Multitype multi;
  Boolean is_resource;
  struct _Property *next;
} *Property;

typedef enum {INITIAL, IN_DESCRIPTION, EXPECT_VALUE, IN_VALUE, SKIP} State;

typedef struct _State {
  State state;
  struct _State *next;
} *StateStack;


static Boolean has_error = False;

static Symbol xml = {"xml", "http://www.w3.org/XML/1998/namespace", NULL};
static SymbolTable symtable = &xml;
static Stack stack = NULL;

static Property properties = NULL; /* List of properties */
static Property last_property = NULL; /* Last of properties list */

static struct _State initial_state = {INITIAL, NULL};
static StateStack state_stack = &initial_state;


/* xml_escape -- replace <, > and & */
static string xml_escape(const string s)
{
  string t;
  int i, j, len;

  assert(s);
  for (i = 0, len = 0; s[i]; i++)
    switch(s[i]) {
    case '<': len += 4; break;
    case '>': len += 4; break;
    case '&': len += 5; break;
    default: len++;
    }
  newarray(t, len + 1);
  for (i = 0, j = 0; s[i]; i++)
    switch (s[i]) {
    case '<': t[j++]='&'; t[j++]='l'; t[j++]='t'; t[j++]=';'; break;
    case '>': t[j++]='&'; t[j++]='g'; t[j++]='t'; t[j++]=';'; break;
    case '&': t[j++]='&'; t[j++]='a'; t[j++]='m'; t[j++]='p'; t[j++]=';'; break;
    default: t[j++] = s[i];
    }
  t[j] = '\0';
  return t;
}

/* xml_unescape -- expand XML entities */
static string xml_unescape(const string s)
{
  string t;
  int i, j;

  newarray(t, strlen(s) + 1);
  for (i = 0, j = 0; s[i];) {
    if (s[i] != '&') {t[j++] = s[i++];}
    else if (strncmp(s + i, "&lt;", 4) == 0) {t[j++] = '<'; i += 4;}
    else if (strncmp(s + i, "&gt;", 4) == 0) {t[j++] = '>'; i += 4;}
    else if (strncmp(s + i, "&amp;", 4) == 0) {t[j++] = '&'; i += 5;}
    else {t[j++] = s[i++];}	/* To do: convert &#...; to UTF-8 */
  }
  t[j] = '\0';
  return t;
}

/* store_property -- add a property at the end of the properties list */
static void store_property(string name, string val, string lang,
			   Multitype multi, Boolean is_resource)
{
  Property h;
  string p;

  new(h);
  h->name = name;
  h->val = val;
  h->lang = lang;
  if (h->lang) for (p = h->lang; *p; p++) *p = tolower(*p);
  h->multi = multi;
  h->is_resource = is_resource;
  h->next = NULL;

  if (last_property == NULL) {	/* List is still empty */
    properties = h;
    last_property = properties;
  } else {
    last_property->next = h;
    last_property = h;
  }
}

/* push_state -- at each open tag, push the current state */
static void push_state(State s)
{
  StateStack h;

  new(h);
  h->state = s;
  h->next = state_stack;
  state_stack = h;
}

/* pop_state -- at each close tag, pop the current state */
static void pop_state(void)
{
  StateStack h;

  if (state_stack == &initial_state) {
    if (! has_error) fprintf(stderr, "%d: too many end tags\n", lineno);
    has_error = True;
    return;
  }
  assert(state_stack);
  h = state_stack;
  state_stack = h->next;
  dispose(h);
}

/* top_state -- return current state */
static State top_state(void)
{
  assert(state_stack);
  return state_stack->state;
}

/* declare_namespaces -- find namespace decls, add them to symbol table */
static void declare_namespaces(pairlist attribs)
{
  Stack h;
  pairlist p;
  Symbol *sym;

  /* Mark the current end of the symbol table */
  new(h);
  h->next = stack;
  h->frame = symtable;
  stack = h;

  /* Scan the attributes for namespace definitions and store them */
  for (p = attribs; p; p = p->next) {
    if (strncmp(p->name, "xmlns", 5) == 0) {
      new(sym);
      sym->prefix = newstring(p->name + (p->name[5] ? 6 : 5));
      sym->uri = xml_unescape(p->value);
      sym->next = symtable;
      symtable = sym;
    }
  }
}

/* pop_namespaces -- unwind the symbol table to previous mark */
static void pop_namespaces(void)
{
  Symbol *h;
  Stack p;

  if (!stack) {
    if (! has_error) fprintf(stderr, "%d: too many end tags\n", lineno);
    has_error = True;
    return;
  }
  /* Remove entries from symbol table chain until last mark */
  while (symtable != stack->frame) {
    h = symtable;
    symtable = symtable->next;
    dispose(h->prefix);
    dispose(h->uri);
    dispose(h);
  }
  /* Pop stack itself */
  p = stack;
  stack = stack->next;
  dispose(p);
}

/* globalname -- expand a prefixed name x:y to uri+y */
static string globalname(string name, Boolean use_default)
{
  string h, prefix, local;
  Symbol *s;

  /* Split the name */
  h = strchr(name, ':');
  if (!h && !use_default) return newstring(name); /* Attrib. without prefix */
  if (h) {
    prefix = newnstring(name, h - name);
    local = h + 1;
  } else {
    prefix = newstring("");
    local = name;
  }
  /* Find the prefix in the symbol table */
  for (s = symtable; s && !eq(prefix, s->prefix); s = s->next) ;

  if (!s && !eq(prefix, "")) {
    fprintf(stderr, "%d: prefix \"%s\" not defined\n", lineno, prefix);
    has_error = True;
    /* To do: do we report anything if the default prefix is undefined? */
  }
  dispose(prefix);
  return strapp(NULL, s ? s->uri : (string)"", local, NULL);
}

/* start_of_local_name -- return start of local name in a global name */
static int start_of_local_name(const string name)
{
  int i;

  assert(name);
  i = strlen(name) - 1;
  /* To do: Test for XML's NameChar instead of isalnum() */
  while (i >= 0 && (isalnum(name[i]) || name[i] == '_' || name[i] == '-'
		    || name[i] == '.')) i--;
  return i + 1;
}

/* langmatch -- check if language lang1 includes lang2 (en includes en-us) */
static Boolean langmatch(const string lang1, const string lang2)
{
  string s = lang1, t = lang2;

  for (; *s && *s == *t; s++, t++);
  return *s == '\0' && (*t == '\0' || *t == '-');
}

/* print_property -- print the value of the property in the given language */
static void print_property(const string property, const string language,
			   const Boolean verbose)
{
  Property p;

  for (p = properties; p; p = p->next) {
    if (fnmatch(property, p->name, 0) == 0 &&
	(!language || (p->lang && langmatch(language, p->lang)))) {
      if (verbose) printf("%s %s ", p->name, p->lang ? p->lang : "");
      printf("%s\n", p->val);
    }
  }
}

/* normalize_list -- remove properties without values, group values together */
static void normalize_list(void)
{
  Property p, q, h;

  /* Remove properties with NULL values (i.e., properties with complex values */
  while (properties && !properties->val) {
    h = properties;
    properties = properties->next;
    dispose(h->name);
    dispose(h->lang);
    dispose(h);
  }
  for (p = properties; p; p = p->next) {
    if (p->next && !p->next->val) {
      h = p->next;
      p->next = p->next->next;
      dispose(h->name);
      dispose(h->lang);
      dispose(h);
    }
  }

  /* Group all values for the same property together. To do: speed? */
  for (p = properties; p && p->next; p = p->next) {
    for (q = p; q->next;) {
      if (!eq(p->name, q->next->name)) {
	q = q->next;
      } else {
	h = q->next;
	q->next = h->next;
	h->next = p->next;
	p->next = h;
	p = p->next;
 	if (q->next == p) q = p;
      }
    }
  }
  if ((last_property = properties))
    while (last_property->next) last_property = last_property->next;
}

/* remove_duplicates -- remove properties that are duplicates of this one */
static void remove_duplicates(Property prop)
{
  Property p, h;

  assert(prop->val);
  assert(properties->val);

  /* Check if first property is a duplicate (apart from language) */
  while (properties && properties != prop &&
	 eq(properties->name, prop->name) && eq (properties->val, prop->val)) {
    h = properties;
    properties = properties->next;
    dispose(h->name);
    dispose(h->lang);
    dispose(h);
  }

  /* Check if other properties are duplicates (apart from langauge) */
  for (p = properties; p; p = p->next) {
    while (p->next && p->next != prop &&
	   eq(p->next->name, prop->name) && eq(p->next->val, prop->val)) {
      h = p->next;
      p->next = p->next->next;
      dispose(h->name);
      dispose(h->lang);
      dispose(h);
    }
  }
}

/* simplify_list -- remove language value that are equal to a default value */
static void simplify_list(void)
{
  Property p;

  normalize_list();		/* Remove NULL values */
  for (p = properties; p; p = p->next)
    /* Be careful that remove_duplicates() does not delete p itself */
    if (p->multi != SEQ &&
	(!p->lang || eq(p->lang, "") || eq(p->lang, "x-default")))
      remove_duplicates(p);
}

/* print_xmp -- print the complete set of properties in XMP syntax */
static void print_xmp(void)
{
  Property p;
  int i, j;

  /* The "begin" pseudo-attribute indicates UTF-8. To do: convert input */
  puts("<?xpacket begin='\357\273\277' id='W5M0MpCehiHzreSzNTczkc9d'?>");
  /*puts("<x:xmpmeta xmlns:x='adobe:ns:meta/'>");*/
  puts(" <rdf:RDF xmlns:rdf='http://www.w3.org/1999/02/22-rdf-syntax-ns#'>");
  puts("  <rdf:Description rdf:about=''>");

  normalize_list();

  for (p = properties; p; p = p->next) {
    j = start_of_local_name(p->name);
    printf("   <%s xmlns='", p->name + j);
    for (i = 0; i < j; i++) putchar(p->name[i]);
    putchar('\'');
    if (p->next && eq(p->name, p->next->name)) {
      puts(">");
      switch (p->multi) {
      case ALT: puts("    <rdf:Alt>"); break;
      case SINGLE:		/* Treat multiple single values as a bag */
      case BAG: puts("    <rdf:Bag>"); break;
      case SEQ: puts("    <rdf:Seq>"); break;
      }
      while (True) {
	printf("     <rdf:li");
	if (p->lang) printf(" xml:lang='%s'", p->lang);
	if (p->is_resource) printf(" rdf:resource='%s'/>\n",xml_escape(p->val));
	else printf(">%s</rdf:li>\n", xml_escape(p->val));
	if (!p->next || !eq(p->name, p->next->name)) break;
	p = p->next;
      }
      switch (p->multi) {
      case ALT: puts("    </rdf:Alt>"); break;
      case SINGLE:
      case BAG: puts("    </rdf:Bag>"); break;
      case SEQ: puts("    </rdf:Seq>"); break;
      }
      printf("   </%s>\n", p->name + j);
    } else {
      if (p->lang) printf(" xml:lang='%s'", p->lang);
      if (p->is_resource) printf(" rdf:resource='%s'/>\n", xml_escape(p->val));
      else printf(">%s</%s>\n", xml_escape(p->val), p->name + j);
    }
  }

  puts("  </rdf:Description>");
  puts(" </rdf:RDF>");
  /*puts("</x:xmpmeta>");*/
  puts("<?xpacket end='w'?>");
}

/* print_sh -- print the complete set of properties in bash shell syntax */
static void print_sh(void)
{
  Property p;
  int i, j, n;

  /* To do: output in current locale's encoding */

  normalize_list();

  for (p = properties, n = 0; p; p = p->next, n++) {
    j = start_of_local_name(p->name);
    printf("xmp_globalname[%d]='%s'\n", n, p->name);
    printf("xmp_localname[%d]='%s'\n", n, p->name + j);
    printf("xmp_namespace[%d]='", n);
    for (i = 0; i < j; i++) putchar(p->name[i]);
    putchar('\''); putchar('\n');
    if (p->lang) printf("xmp_language[%d]='%s'\n", n, p->lang);
    if (p->is_resource) printf("xmp_is_resource[%d]=1\n", n);
    /* To do: escape \n and ' */
    printf("xmp_value[%d]='%s'\n", n, p->val);
    printf("xmp_multi[%d]='%s'\n", n, p->multi == SINGLE ? "Single" :
	   p->multi == ALT ? "Alt" : p->multi == BAG ? "Bag" : "Seq");
  }
}

/* delete -- delete a property or a property's value in a specific language */
static void delete(string prop, string lang)
{
  Property p, h;

  /* Check the first property in the list */
  while (properties && fnmatch(prop, properties->name, 0) == 0 &&
	 (!lang ||
	  (properties->lang && fnmatch(lang, properties->lang, 0) == 0))) {
    h = properties;
    properties = properties->next;
    dispose(h->name);
    dispose(h->val);
    dispose(h->lang);
    dispose(h);
    if (!properties) last_property = NULL;
  }

  /* Check the rest of the property list */
  for (p = properties; p && p->next;) {
    if (fnmatch(prop, p->next->name, 0) == 0 &&
	(!lang || (p->next->lang && fnmatch(lang, p->next->lang, 0) == 0))) {
      h = p->next;
      if (last_property == h) last_property = p;
      p->next = p->next->next;
      dispose(h->name);
      dispose(h->val);
      dispose(h->lang);
      dispose(h);
    } else {
      p = p->next;
    }
  }
}

/* handle_properties_in_attribs -- look for properties in attribute list */
static void handle_properties_in_attribs(pairlist attribs)
{
  pairlist p;
  string s;

  for (p = attribs; p; p = p->next) {
    if (strncmp(p->name, "xmlns", 5) != 0) { /* It's not a namespace */
      s = globalname(p->name, False);
      if (strncmp(s, "http://www.w3.org/1999/02/22-rdf-syntax-ns#", 43) != 0) {
	store_property(s, xml_unescape(p->value), NULL, SINGLE, False);
      } else {
	dispose(s);		/* It's an RDF attribute, probably "about" */
      }
    }
  }
}

/* has_about_attrib -- check if there is an rdf:about attribute in the list */
static Boolean has_about_attrib(pairlist attribs)
{
  pairlist p;
  Boolean found;
  string s;

  for (p = attribs, found= False; p && !found; p = p->next) {
    if (strncmp(p->name, "xmlns", 5) == 0) continue;
    if (!strstr(p->name, "about")) continue;
    s = globalname(p->name, False);
    found = eq(s, "http://www.w3.org/1999/02/22-rdf-syntax-ns#about")
      || eq(s, "about");	/* "about" w/o prefix is deprecated, but used */
    dispose(s);
  }
  return found;
}

/* parsetype_is_resource -- check for rdf:parseType="Resource" */
static Boolean parsetype_is_resource(pairlist attribs)
{
  pairlist p;
  Boolean found;
  string s;

  for (p = attribs, found= False; p && !found; p = p->next) {
    if (strncmp(p->name, "xmlns", 5) == 0) continue;
    if (!strstr(p->name, "parseType")) continue;
    s = globalname(p->name, False);
    found = eq(s, "http://www.w3.org/1999/02/22-rdf-syntax-ns#parseType")
      && eq(p->value, "Resource");
    dispose(s);
  }
  return found;
}

/* find_lang -- look for xml:lang in the list of attributes and return value */
static string find_lang(pairlist attribs)
{
  pairlist p;

  for (p = attribs; p; p = p->next)
    if (eq(p->name, "xml:lang")) return xml_unescape(p->value);
  return NULL;
}

/* find_resource -- look for rdf:resource in attributes and return value */
static string find_resource(pairlist attribs)
{
  pairlist p;
  string s, resource;

  for (p = attribs, resource = NULL; !resource && p; p = p->next) {
    if (strncmp(p->name, "xmlns", 5) == 0) continue;
    if (!strstr(p->name, "resource")) continue;
    s = globalname(p->name, False);
    if (eq(s, "http://www.w3.org/1999/02/22-rdf-syntax-ns#resource"))
      resource = xml_unescape(p->value);
    dispose(s);
  }
  return resource;
}


/* handle_error -- called when a parse error occurred */
void handle_error(void *clientdata, const string s, int lineno)
{
  fprintf(stderr, "%d: %s\n", lineno, s);
  has_error = True;
}

/* start -- called before the first event is reported */
void* start(void)
{
  return NULL;
}
  
/* end -- called after the last event is reported */
void end(void *clientdata)
{
  /* skip */
}

/* handle_comment -- called after a comment is parsed */
void handle_comment(void *clientdata, string commenttext)
{
  /* skip */
}

/* handle_text -- called after a text chunk is parsed */
void handle_text(void *clientdata, string text)
{
  string s;

  if (top_state() == IN_VALUE) {
    assert(last_property);
    s = xml_unescape(text);
    last_property->val = strapp(&last_property->val, s, NULL);
    dispose(s);
  } else if (top_state() == EXPECT_VALUE) {
    assert(text);
    for (; isspace(*text); text++); /* Skip white space */
    if (*text) {
      s = xml_unescape(text);
      last_property->val = strapp(&last_property->val, s, NULL);
      dispose(s);
      pop_state();
      push_state(IN_VALUE);	/* Next time don't skip white space */
    }
  }
}

/* handle_decl -- called after a declaration is parsed */
void handle_decl(void *clientdata, string gi, string fpi, string url)
{
  /* skip */
}

/* handle_pi -- called after a PI is parsed */
void handle_pi(void *clientdata, string pi_text)
{
  /* skip */
}

/* handle_starttag -- called after a start tag is parsed */
void handle_starttag(void *clientdata, string name, pairlist attribs)
{
  string s, lang, resource;

  declare_namespaces(attribs);
  s = globalname(name, True);
  if (top_state() == SKIP) {
    push_state(SKIP);		/* Keep skipping */
  } else if (eq(s, "http://www.w3.org/1999/02/22-rdf-syntax-ns#RDF"))
    push_state(top_state());	/* Skip top-level element */
  else if (eq(s, "http://www.w3.org/1999/02/22-rdf-syntax-ns#Description")) {
    if (has_about_attrib(attribs)) {
      /* <rdf:Description rdf:about=""> */
      push_state(IN_DESCRIPTION);
      handle_properties_in_attribs(attribs);
    } else {
      push_state(top_state());
    }
  } else if (eq(s, "http://www.w3.org/1999/02/22-rdf-syntax-ns#value")) {
    /* <rdf:value> */
    if (top_state() == EXPECT_VALUE) {
      push_state(IN_VALUE);	/* Initial spaces are not skipped */
      if ((lang = find_lang(attribs))) {
	assert(last_property);
	dispose(last_property->lang); /* Set/replace the language */
	last_property->lang = lang;
      }
    } else {
      push_state(top_state());
    }
  } else if (eq(s, "http://www.w3.org/1999/02/22-rdf-syntax-ns#Bag")) {
    last_property->multi = BAG;
    push_state(top_state());
  } else if (eq(s, "http://www.w3.org/1999/02/22-rdf-syntax-ns#Alt")) {
    last_property->multi = ALT;
    push_state(top_state());
  } else if (eq(s, "http://www.w3.org/1999/02/22-rdf-syntax-ns#Seq")) {
    last_property->multi = SEQ;
    push_state(top_state());
  } else if (eq(s, "http://www.w3.org/1999/02/22-rdf-syntax-ns#li")) {
    if (parsetype_is_resource(attribs)) {
      push_state(SKIP);		/* Skip resources that are not strings */
    } else if (top_state() == EXPECT_VALUE) {
      if (last_property->val) 	/* This is not the first <li> */
	store_property(newstring(last_property->name), NULL, NULL,
		       last_property->multi, False);
      if ((lang = find_lang(attribs))) { /* Set language */
	dispose(last_property->lang);
	last_property->lang = lang;
      }
      if ((resource = find_resource(attribs))) {
	last_property->val = resource;
	last_property->is_resource = True;
	push_state(SKIP);	/* Don't expect another value */
      } else {
	last_property->is_resource = False;
	push_state(EXPECT_VALUE); /* Still expect a value */
      }
    } else if (top_state() == SKIP) {
      push_state(SKIP);
    } else {			/* Structure of file is wrong? */
      fprintf(stderr, "%d: unexpected <li> ignored\n", lineno);
      has_error = True;
      push_state(SKIP);
    }
  } else if (top_state() == IN_DESCRIPTION && !parsetype_is_resource(attribs)) {
    /* Must be a property name */
    if ((resource = find_resource(attribs))) {
      store_property(newstring(s), resource, find_lang(attribs), SINGLE, True);
      push_state(SKIP);		/* Don't expect another value */
    } else {
      store_property(newstring(s), NULL, find_lang(attribs), SINGLE, False);
      push_state(EXPECT_VALUE); /* Still expect a value */
    }
  } else if (top_state() == INITIAL) {
    push_state(INITIAL);	/* Not yet inside a Description */
  } else {
    push_state(SKIP);		/* Inside blank node or complicated things */
  }
  dispose(s);
  dispose(name);
}

/* handle_endtag -- called after an endtag is parsed (name may be "") */
void handle_endtag(void *clientdata, string name)
{
  pop_namespaces();
  pop_state();
}

/* handle_emptytag -- called after an empty tag is parsed */
void handle_emptytag(void *clientdata, string name, pairlist attribs)
{
  handle_starttag(clientdata, name, attribs);
  handle_endtag(clientdata, NULL);
}


/* usage -- print usage message and exit */
static void usage(string prog)
{
  errexit("%s version %s\n\
Usage: %s [-v] [-l language] property [file]\n\
       %s -w [-r] [-l language] property value [file]\n\
       %s -d [-l language] property [file]\n\
       %s -s [file]\n",
	  prog, VERSION, prog, prog, prog, prog);
}

/* main -- main body */
int main(int argc, char *argv[])
{
  string property = NULL, value = NULL, language = NULL;
  enum {PRINT, WRITE, SHELL, DELETE, CLEAN} task = PRINT;
  Boolean is_resource = False, verbose = False;
  int c;

  /* Bind the parser callback routines to our handlers */
  set_error_handler(handle_error);
  set_start_handler(start);
  set_end_handler(end);
  set_comment_handler(handle_comment);
  set_text_handler(handle_text);
  set_decl_handler(handle_decl);
  set_pi_handler(handle_pi);
  set_starttag_handler(handle_starttag);
  set_emptytag_handler(handle_emptytag);
  set_endtag_handler(handle_endtag);

  while ((c = getopt(argc, argv, "cvsrwdl:")) != -1)
    switch (c) {
    case 'w': if (task == PRINT) task = WRITE; else usage(argv[0]); break;
    case 'l': language = optarg; break;
    case 'r': is_resource = True; break;
    case 'd': if (task == PRINT) task = DELETE; else usage(argv[0]); break;
    case 's': if (task == PRINT) task = SHELL; else usage(argv[0]); break;
    case 'v': if (task == PRINT) verbose = True; else usage(argv[0]); break;
    case 'c': if (task == PRINT) task = CLEAN; else usage(argv[0]); break;
    default: exit(1); /* usage(argv[0]); */
    }
  if (task != SHELL && task != CLEAN) {
    if (argc != optind) property = argv[optind++]; else usage(argv[0]);
  }
  if (task == WRITE) {
    if (argc != optind) value = argv[optind++]; else usage(argv[0]);
  }
  if (optind == argc) yyin = stdin;
  else if (optind == argc - 1) yyin = fopen(argv[optind], "r");
  else usage(argv[0]);
  if (yyin == NULL) {perror(argv[optind]); exit(2);}
  if (yyparse() != 0) {exit(3);}

  switch (task) {
  case WRITE:
    assert(property);
    assert(value);
    store_property(property, value, language ? language : NULL, SINGLE,
		   is_resource);
    print_xmp();
    break;
  case SHELL:
    print_sh();
    break;
  case PRINT:
    assert(property);
    print_property(property, language, verbose);
    break;
  case DELETE:
    assert(property);
    delete(property, language);
    print_xmp();
    break;
  case CLEAN:
    simplify_list();
    print_xmp();
  default:
    assert("Cannot happen!");
  }

  return has_error ? 1 : 0;
}
